#include <cctype>
#include <cmath>
#include <fstream>
#include <iostream>
#include <string>
#include "console.h"
#include "simpio.h"
#include "queue.h"
#include "map.h"
#include "strlib.h"
#include "filelib.h"
#include "random.h"

using namespace std;

// Function  declarations

void printWelcomeMessage();
void openTextFile(ifstream &inputFile);
void setNgram(int &nGramVal);
void makeRandomString(Vector<string> vec,string &word);
void displayVectorOfStrings(Vector<string> vec);
Map<Vector<string>,Vector<string>> generateMap(ifstream &inputFile, int n);
void printText(int numOfRandWords, Map<Vector<string>, Vector<string>> &map);

// main() function

int main() {

    printWelcomeMessage();
    ifstream inputFile;
    openTextFile(inputFile);

    int nGramVal;
    setNgram(nGramVal);

    Map<Vector<string>, Vector<string>>  map = generateMap(inputFile, nGramVal);
    cout << endl;

    while (true) {
        int numOfRandWords = getInteger("# of random words to generate (0 to quit)?");

        while (numOfRandWords < nGramVal && numOfRandWords!=0) {
            cout << "Must be at least " << nGramVal << " words." << endl;
            numOfRandWords = getInteger("# of random words to generate (0 to quit)?");
        }
        while(numOfRandWords<4 && numOfRandWords!=0) {
            cout<<"Must be at least 4 words."<<endl;;
            numOfRandWords = getInteger("# of random words to generate (0 to quit)?");
        }

        if (numOfRandWords == 0){
            break;
        }
        printText(numOfRandWords, map);
        cout << endl;
    }

    cout << "Exiting." << endl;
    return 0;
}

// This function prints the welcome message

void printWelcomeMessage() {
    cout << "Welcome to N-Grams Random Text Generator." << endl;
    cout << "This program makes random text based on a document." << endl;
    cout << "Give me an input file and an 'N' value for groups" << endl;
    cout << "of words, and I'll create random text for you."<< endl << endl;
}

// This function checks for the validity of nGramVal

void setNgram(int &nGramVal)  {
    nGramVal = getInteger("Value of N?","Illegal integer format.");
    while (nGramVal < 2) {
        cout << "N must be 2 or greater." << endl;
        nGramVal = getInteger("Value of N?","Illegal integer format");
    }
}

// This function opens the text file provided by the user

void openTextFile(ifstream &inputFile) {
        string filename = getLine("Input file name? ");
        while(!fileExists(filename))    {
            cout<<"Unable to open that file. Try again."<<endl;
            filename=getLine("Input file name? ");
        }
        openFile(inputFile,filename);
}

// This function stores the value of a vector of random index in word whic has been passed by reference

void makeRandomString(Vector<string> vec,string &word) {
    int randomInt = randomInteger(0, vec.size() - 1);
    word = vec[randomInt];
}

// This vector prints out a vector of strings

void displayVectorOfStrings(Vector<string> vec) {
    for (string word : vec) {
        cout << word << " ";
    }
}

// This function converts the file data into a map with keys and corresponding values

Map<Vector<string>,Vector<string>> generateMap(ifstream &inputFile, int nGramVal) {
    Map<Vector<string>,Vector<string>> map;
    Vector<string> wrapWindow;
    Vector<string> window;

    int i = 1;
    while (i++ < nGramVal) {
        string nextWord;
        inputFile >> nextWord;
        wrapWindow.add(nextWord);
    }
    window = wrapWindow;

    while (true) {                      // The while loop remains true as long as the last word of the file is not attained
        string nextWord;
        inputFile >> nextWord;
        if(inputFile.fail()){
            break;
        }
        if (map.containsKey(window)) {
            map[window] += nextWord;
        } else {
            map[window] = {nextWord};
        }
        window.add(nextWord);
        window.remove(0);
    }
    for (string s : wrapWindow) {       // This for loop helps in wrapping the text using the wrapWindow(which contains the first N-1 words needed for wrapping)
        map[window]+= s;
        window.add(s);
        window.remove(0);
    }
    inputFile.close();
    return map;
}

// This function prints random text.

void printText(int numOfRandWords, Map<Vector<string>, Vector<string>> &map) {
    Vector<Vector<string>> keys = map.keys();
    int randomInt = randomInteger(0, keys.size() - 1);
    Vector<string> keyWindow = keys[randomInt];
    int size = keyWindow.size();

    Vector<string> text;
    text += keyWindow;

    for (int i = size; i < numOfRandWords; ++i) {
        string nextWord;
        makeRandomString(map[keyWindow],nextWord);
        text.add(nextWord);
        keyWindow.add(nextWord);
        keyWindow.remove(0);
    }
    cout << "... ";
    displayVectorOfStrings(text);
    cout << "..." << endl;
}
